''*****************************************
''*   MIDI file player for Ahles SID Emu  *
''*****************************************
'' Author:   Andy Schenk, Insonix
'' Date:     Dec 2009

'' Needs an SD card connected, and the MIDI files stored on the card.
'' Because the files are played directly from the card, only the MidiFile Format 0
'' is supported (all Tracks mixed down to one).


CON
  _clkmode        = xtal1 + pll16x
  _xinfreq        = 5_000_000

  SD_basepin      = 0       '<-- SD card
  Left            = 10      '<-- Audio Out pins
  Right           = 11
        
VAR    
  long  time, tq, ts, us
  byte  filename[16]

OBJ
  sd    : "fsrw"
  SID   : "SIDcog"   'or "SIDCogRes"


DAT
playlist
                byte    "POPCORN0.MID",0
listend

      
PUB Main

  SID.start                              'Start the emulated SID chip in one cog 
                    'Lp,   Bp,   Hp    
  SID.setFilterType(true,false,false)    'Enable lowpass filter

                    'Ch1, Ch2, Ch3      
  SID.setFilterMask(true,false,false)    'Enable filter on the bass channel (channel1)

  SID.setWaveform(0,2)                   'Set waveform type on channel1 to Square wave
  SID.setPWM(0,1928)                     'Set the pulse width on channel1 to 47:53
  SID.setADSR(0,1,7,1,9)                 'Set Envelope on channel3  

  SID.setWaveform(1,0)                   'Set waveform type on channel2 to Triangle  
  SID.setADSR(1,1,7,9,4)                 'Set Envelope on channel2  

  SID.setWaveform(2,1)                   'Set waveform type on channel3 to Saw wave
  SID.setADSR(2,5,9,0,8)                 'Set Envelope on channel3  

  SID.setVolume(15)                      'Set volume to max
  SID.setCutoff(400)                     'Filter

  ifnot sd.mount(SD_basepin)                            'init SD card
    filename := @playlist                               'start Play List
    repeat
      waitcnt(clkfreq*2+cnt)
      midiplay(filename)                                'play the file
      filename += strsize(filename)+1
    until filename => @listend                          'next file

  repeat
    

PUB midiplay(name)  | lg, dt, b, v, st, d1, d2, ch
  us := clkfreq / 1_000_000                             'ticks / microsecond
  b := st := d1 := d2 := v := 0
  ifnot sd.popen(name,"r")
    sd.pread(@v,4)
    if v == "M" + "T"<<8 + "h"<<16 + "d"<<24            'Header ID
      sd.pread(@v,4)                                    'Header Len
      v~
      sd.pread(@v,2)                                    'Format
      if v <> 0
'        disp.str(string("only Format 0 supported"))
      else
        sd.pread(@v,2)                                  'Tracks
        sd.pread(@v,2)                                  'Ticks/Quarter
        tq := v>>8 + (v&255)<<8
        ts := 500_000 / tq                              'default Tempo 120 BPM
        sd.pread(@v,4)                                  '"MTrk"
        sd.pread(@v,4)                                  'Track length
        lg := v>>24 + (v>>16&255)<<8 + (v>>8&255)<<16
        time := cnt
        repeat
          sd.pread(@b,1)                                'get delta time
          lg--
          dt := b & $7F
          repeat while b>127
            sd.pread(@b,1)
            lg--
            dt := dt<<7 + (b & $7F)
          time := time + dt*us*ts
          repeat until time-cnt < 0                     'wait deltatime
          b~
          sd.pread(@b,1)                                'MIDI byte
          if b > 127
            st := b                                     'new status byte
          elseif st==$FF
            st~
          ch := st & $0F
          d1~
          d2~
          case st & $F0                                 'decode MIDI Event
            $90: sd.pread(@d1,1)                        'Note 
                 sd.pread(@d2,1)                        'Velocity
                 lg -= 3
                 if d2>0
                   noteOn(d1,ch,d2)               'Note On
                 else
                   noteOff(d1,ch)                 'Note Off if Vel=0
            $80: sd.pread(@d1,1)                        'Note Off 
                 sd.pread(@d2,1)                        'Velocity
                 lg -= 3
                 noteOff(d1,ch)
            $A0: sd.pread(@d1,1)                        'Poly AfterTouch 
                 sd.pread(@d2,1)                        '(not supported)
                 lg -= 3
            $C0: sd.pread(@d1,1)                        'Prg Change 
                 lg -= 2
                 prgChange(d1,ch)
            $D0: sd.pread(@d1,1)                        'Mono AfterTouch 
                 lg -= 2                                '(not supported)
            $B0: sd.pread(@d1,1)                        'Controller (nr) 
                 sd.pread(@d2,1)                        'value
                 lg -= 3
                 if d1==7
                   volContr(d2,ch)                '7=Vol
            $E0: sd.pread(@d1,1)                        'Pitch Bender 
                 sd.pread(@d2,1)                        '(not supported!)
                 lg -= 3
            $F0: sd.pread(@d1,1)                        'Meta Event
                 v~
                 if st==$F0 or st==$F7                  'SysEx
                   repeat
                     sd.pread(@v,1)
                     lg--
                     d1--
                   until v==$F7 or b=< 0
                   d1~
                 else
                   lg -= 2
                   sd.pread(@b,1)                       'Len
                   repeat while b>0                     'read meta data
                     sd.pread(@d2,1)
                     v := v<<8 + d2
                     b--
                     lg--
                   if d1==81                            'Tempo
                     ts := v / tq
            0:   sd.pread(@d1,1)                        'Meta Event Running status
                 lg -= 2
                 sd.pread(@b,1)                         'Len
                 repeat while b>0                       'read meta data
                   sd.pread(@d2,1)
                   v := v<<8 + d2
                   b--
                   lg--
            
        until lg < 1
    else
'      disp.str(string("No MIDI file"))
    sd.pclose
  else
'    disp.str(string("File not found"))

PRI noteOn(note,chl,vel) | octave, sidch
    octave := note/12
    note -= octave*12  'CH 1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16   MIDI Channel
    sidch := lookupz(chl : 0, 1, 2, 3, 3, 1, 3, 2, 0, 3, 3, 3, 3, 3, 3, 3) '=SID Channel
'    sidch := chl//3
    if sidch < 3
      SID.noteOn(sidch, noteTable[note]>>(8-octave)) 

PRI noteOff(key,chl)
  SID.noteOff(chl//3)

PRI prgChange(num,chl)

PRI volContr(vo,chl)
'  SID.setVolume(vo>>4)


DAT
noteTable word 16350, 17320, 18350, 19450, 20600, 21830, 23120, 24500, 25960, 27500, 29140, 30870
     
{{
 ---------------------------------------------------------------------------
                Terms of use: MIT License                                   
 ---------------------------------------------------------------------------
   Permission is hereby granted, free of charge, to any person obtaining a  
  copy of this software and associated documentation files (the "Software"),
  to deal in the Software without restriction, including without limitation 
  the rights to use, copy, modify, merge, publish, distribute, sublicense,  
    and/or sell copies of the Software, and to permit persons to whom the   
    Software is furnished to do so, subject to the following conditions:    
                                                                            
   The above copyright notice and this permission notice shall be included  
           in all copies or substantial portions of the Software.           
                                                                            
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER   
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING   
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER     
                       DEALINGS IN THE SOFTWARE.                            
 ---------------------------------------------------------------------------
}}  